/*
 * LCD_Config v2.0 beta
 * --------------------
 *
 * allows configuration of the MultiWii software via a serial connection.
 * requires a LCD display and at least four buttons (left, right, up, down).
 *
 * Options:
 *   Logging functionality (requires a SD card).
 *   Status LEDs (three LEDs, two are sufficient without logging)
 *   Setting of LCD Backlight intensity
 *   Use of ShiftRegLCD lib instead of LiquidChrystal lib
 *   Hardware buttons instead of buttons connected via cascading resistors
 *   Graphic output (simple character graphic for angles and servo signals)
 *   MinMax mode (store und display minimum/maximum values of certain fields) 
 *   [Test] (for standalone testing without a MWC board connected)
 *
 * IMPORTANT: undefining the LOGGING function does not work correctly. When commenting
 * out the line "#include <SD.h>", the image size decreases by about 4k. Recommended
 * in case you are targeting a ATMega168.
 *
 * Authors: yacco, calli
 *
 * changes to 1.25: 
 *   display values of gyro and acc changed to signed
 *   show minimum/maximum values of sensor data and servo signals
 * changes to 1.28:
 *   wait state after startup, no serial communication during initialization of the MWC
 *   removed min/max values
 *   added LED status signals
 * changes to 1.30:
 *   bugfix in playback routine, added timeout in playback
 * changes to 1.31:
 *   minor bugfixes, added setting of backlight
 * changes to 1.32:
 *   removed KEY_SELECT column in next_state, backlight and waitForInit optional
 * changes to 1.33:
 *   check trailing 'A' char in serial communication to detect frame errors, 
 *   optimization of format routines (merged two routines, invalid values are displayed as ###)
 *   bugfix in logging (logMode=-1 was stored in flash, so logging was permanently disabled)
 *   KEY_LEFT cancels reading when no serial connection is present (important to reset wrong settings!) 
 * changes to 1.5
 *   reimplemented Min/Max-Mode (found it useful again), 
 *   fixed compiling problems for certain combinations of options 
 * changes to 1.6
 *   adaptation to MWC version 1.preter6 (new config "Flying Wing", new data format)
 *   changed status display (state 0 - Config)
 *   bug 'logMode=-1' finally solved
 *   Version has been set to 2.0 to disambiguate between MWC version and LCD_Conf version.
 *
 * Graphic Mode is toggled either by pressing the SELECT key or by pressing the RIGHT key longer
 * than 0.5 sec. Min/Max Mode is toggled by pressing the RIGHT key shorter than 0.5 sec.
 
 * overall footprint (on an Arduino Uno, using the LiquidCrystal lib and cascading buttons): 
 *  ~27800 bytes with all options enabled 
 *  ~12500 bytes with all options disabled (and //#include <SD.h>)
 * 
 * -----
 * TO DO: show pan-tilt servos, ...
 */

// ----- begin of configuration parameters -----

// use ShiftRegLCD lib instead of LiquidCrystal lib
// [footprint: decreases ~200 bytes]
//#define SRLCD     

// use Hardwarekeys connected to Ports
// [footprint: ~120 bytes]
//#define HWKEYS

// include logging to SD/SDHC card (formatted to FAT16 or FAT32) via SPI
// [footprint: ~12 kbytes]
#define LOGGING
#define CHIP_SELECT 10

// include routines to show some output as simple character graphic (for servo signals and angles) 
// [footprint: ~600 bytes]
#define GRAPHIC

// track minimum and maximum values of certain values
// [footprint: ~200 bytes]
#define MINMAX

// include routines to show the state of the application with three leds 
// (red: error, yellow: busy/waiting for input, green: logging)
// [footprint: ~300 bytes]
#define STATUS_LEDS

#define RED_LED      A1
#define YELLOW_LED   A3
#define GREEN_LED    A5

#ifdef STATUS_LEDS
  #define green(a)   digitalWrite( GREEN_LED, a)
  #define yellow(a)  digitalWrite(YELLOW_LED, a)
  #define red(a)     digitalWrite(   RED_LED, a)
#else
  #define green(a)
  #define yellow(a)
  #define red(a)
#endif

// if defined, communication to the MWC board is blocked after startup until the user presses a key 
// [footprint: ~120 bytes]
#define WAIT_FOR_INIT

// if defined, allows dimming of LCD backlight intensity
// [footprint: ~640 bytes]
#define BACKLIGHT

#ifdef BACKLIGHT 
  #define setBacklight()    analogWrite(10, backlight)
#else
  #define setBacklight()
#endif

// simulates a serial connection, SELECT key in state "ESC Out" changes the configuration
//#define TEST      

// ----------------

// Number of hardbuttons:
#define NUM_KEYS     5

// Identifiers for the five hardbuttons:
#define KEY_RIGHT    0
#define KEY_UP       1
#define KEY_DOWN     2
#define KEY_LEFT     3
#define KEY_SELECT   4
#define STATUS       5 // Last column in the transition table is reserved for status flags

// Input pins of the hardbuttons: (used when HWKEYS is defined)
#define HW_RIGHT    10
#define HW_UP        9
#define HW_DOWN     11
#define HW_LEFT      8
#define HW_SELECT    6

// RC Channel Aliases:
#define ROLL         0
#define PITCH        1
#define YAW          2
#define THROTTLE     3
#define AUX1         4

// ----- end of configuration parameters -----

// States:
#ifdef LOGGING
  #define NUM_STATES  46
#else
  #define NUM_STATES  42
#endif

// Number of Configurations: (Unknown, Tri, Quad+, QuadX, Bi, Gimbal, Y6, Hex, Flying Wing)
#define NUM_CONFIGS  9

// Speed of the serial interface. 115200 kbit/s is the best option for a USB connection.
#define SERIAL_COM_SPEED 115200

// ----------------

// for persistence of SW Version, Logging Mode, LogDelay and Profiles
#include <EEPROM.h>

// defines the PROGMEM macro to move constant variables from SRAM to flash memory
#include <avr/pgmspace.h> 

// ----------------

// Workaround for "insertion of forward declarations in excluded blocks" problem
void setup();     

// ----------------

#ifndef SRLCD
  #include <LiquidCrystal.h>
  LiquidCrystal lcd = LiquidCrystal(8, 9, 4, 5, 6, 7); // rs, enable, d4, d5, d6, d7
#else
  #include <ShiftRegLCD.h>
  ShiftRegLCD lcd(3, 2, 4, 2);        // (Datapin, Clockpin, Enablepin or TWO_WIRE [, Lines [, Font]]])
#endif

// ----------------

#ifdef LOGGING
  #include <SD.h>                     // ***** comment this line when LOGGING is not required - saves 4k of flash *****
  static File logFile;
  static char fileName[] = "000.log"; // LogFiles consists of a three digit number (000..999) with extension ".log"
#endif

// ----------------

PROGMEM prog_char intro       [] = "                LCD Config 2.0 \342                ";
PROGMEM prog_char emptyLine   [] = "                ";
PROGMEM prog_char reading     [] = " Reading...     ";
PROGMEM prog_char writing     [] = " Writing...     ";
PROGMEM prog_char versMsg     [] = "Select Version: ";
PROGMEM prog_char readMsg     [] = " Read 00 Bytes  ";
PROGMEM prog_char writeMsg    [] = " Wrote 00 Bytes ";
PROGMEM prog_char cancelRead  [] = "Reading canceled";
PROGMEM prog_char frameError  [] = "Inv. FrameLength";
PROGMEM prog_char timeoutError[] = " Timeout Error  ";

PROGMEM prog_char levelMsg    [] = " LevMod:-";
PROGMEM prog_char accMsg      [] = "Acc";
PROGMEM prog_char statusMsg   [] = "Nck  Baro  Mag  ";
PROGMEM prog_char yesNoMsg    [] = " \177: no, \176: yes  ";

PROGMEM prog_char haltMsg1    [] = " Program halted ";
PROGMEM prog_char haltMsg2    [] = " Press 'RESET'  ";
PROGMEM prog_char resetMsg    [] = "! Reset Values !";

#ifdef WAIT_FOR_INIT
PROGMEM prog_char waitMsg1    [] = "Wait for Init...";
PROGMEM prog_char waitMsg2    [] = " Press any key  ";
#endif

#ifdef BACKLIGHT
PROGMEM prog_char bLightMsg   [] = "LCD Backlight:  ";
#endif

#ifdef LOGGING
PROGMEM prog_char logMsg      [] = " Logging...     ";
PROGMEM prog_char noCard      [] = "No SD Card found";
PROGMEM prog_char fileError   [] = "Open File failed";
PROGMEM prog_char fileDeleted [] = "Logfile deleted ";
PROGMEM prog_char deleteFailed[] = "Unable to delete";
PROGMEM prog_char haltReqMsg  [] = " Halt Program?  ";
PROGMEM prog_char logEnbMsg   [] = " Logging enabled";
PROGMEM prog_char stopLogMsg  [] = " Stop Logging?  ";
PROGMEM prog_char playMsg1    [] = " Play Logfile?  ";
PROGMEM prog_char playMsg3    [] = "Select Logfile: ";
PROGMEM prog_char playMsg4    [] = "Play Logfile:   ";
PROGMEM prog_char fileMsg     [] = "File: ";
PROGMEM prog_char fileSuff    [] = "000 #00000";
#endif

#ifdef MINMAX
PROGMEM prog_char minMaxMsg   [] = "Mode:";
PROGMEM prog_char minMaxDsc[][4] = { "Act", "Min", "Max" };
#endif

#ifdef GRAPHIC
PROGMEM prog_char graphicMsg  [] = "Graphic:";
PROGMEM prog_char onOffMsg [][4] = { "OFF", "ON " };
#endif

// ----------------

PROGMEM prog_char versions[][10] = { 
  "1.5      ", 
  "1.prebis6", 
  "1.preter6",
};

// number of bytes per frame send from the MultiWii board, dependant on version
static uint8_t frameLength[] = { 
  65, 
  70, 
  74,
};

// configuration ids
PROGMEM prog_char configs[NUM_CONFIGS][8] = { 
  "Unknown", 
  "    Tri", 
  "  Quad+", 
  "  QuadX", 
  "     Bi", 
  " Gimbal", 
  "     Y6", 
  "    Hex", 
  "Fl.Wing",
};

// ----------------

//State messages
PROGMEM prog_char msgs[NUM_STATES][17] = {
//   0123456789012345 
    "1  Config       ", // 0
    "2  P-I-D        ", // 1
    "3  Rates        ", // 2
    "4  Sensors      ", // 3
    "5  Receiver In  ", // 4
    "6  ESC/Servo Out", // 5
    
    "Roll  P   I   D ", // 6
    "Pitch P   I   D ", // 7
    "Yaw   P   I   D ", // 8

    "D/R   Expo   ALS", // 9
    "R/P   Yaw   Thro", // 10
    
    "Gyr: Rol Pit Yaw", // 11
    "Acc: Rol Pit Z  ", // 12
    "Rho   Phi  Cycle", // 13
    
    "Rec: Rol Pit Yaw", // 14
    "Rec:  Thro  Mode", // 15

    "Lf: 000  Ri: 000", // 16 Tri
    "Ru: 000  Re: 000", // 17
    
    "Lf: 000  Ri: 000", // 18 Quad+
    "Fr: 000  Re: 000", // 19

    "FL: 000  FR: 000", // 20 QuadX
    "RL: 000  RR: 000", // 21

    "M1: 000  M2: 000", // 22 Bi
    "S1: 000  S2: 000", // 23

    " Roll:  000     ", // 24 Gimbal
    "Pitch:  000     ", // 25

    "Up 000  000  000", // 26 Y6 (upper - left rear right)
    "Dn 000  000  000", // 27    (lower - left rear right)

    "Fr 000  000  000", // 28 Hex (front - left mid right)
    "Rr 000  000  000", // 29     (rear  - left mid right)
    
    "Alt   Head      ", // 30
    
    "1  Config       ", // 31 - Acc, Bar, Mag
    
    "7  Profiles     ", // 32
    "8  Settings     ", // 33

    " Load Profile   ", // 34
    " Save Profile   ", // 35
    "Clear Profile   ", // 36

    " Load Profile:  ", // 37
    " Save Profile:  ", // 38
    "Clear Profile:  ", // 39

    "S1: 000  S2: 000", // 40 Flying Wing (servo 0, 1)
    "S3: 000  S4: 000", // 41             (servo 2, 3)

#ifdef LOGGING
    "9  Logging      ", // 42
    "Start Logging   ", // 43
    "Delete Logfile  ", // 44
    "Delay Time      ", // 45
#endif
};

#ifdef LOGGING
  #define UP0 42
  #define DN0 42
#else
  #define UP0 33
  #define DN0 0
#endif

// state transition table, values above 80 are pseudo-states
PROGMEM prog_char nextState[NUM_STATES][NUM_KEYS] = {
  // Right Up  Down Left  Status      Status Flag: 1=Refresh Data, 2=2-Line Mode 
  {   85, UP0,   1,  31,   1, }, //  0 - Config
  {    6,   0,   2,   0,   0, }, //  1 - PID
  {    9,   1,   3,   0,   0, }, //  2 - Rates
  {   11,   2,   4,   0,   0, }, //  3 - Sensors
  {   14,   3,   5,   0,   0, }, //  4 - Rec In
  {   99,   4,  32,   0,   0, }, //  5 - ESc Out

  {   80,   8,   7,   1,   0, }, //  6 - PID Roll
  {   81,   6,   8,   1,   0, }, //  7 - PID Pitch
  {   82,   7,   6,   1,   0, }, //  8 - PID Yaw

  {   83,  10,  10,   2,   0, }, //  9 - DR, Expo, AutoLevel Strength
  {   84,   9,   9,   2,   0, }, // 10 - R/P Rate, Yaw Rate, Throttle PID

  {   93,  30,  12,   3,   1, }, // 11 - Gyro
  {   93,  11,  13,   3,   1, }, // 12 - Acc
  {   93,  12,  30,   3,   1, }, // 13 - Angle, Cycle time

  {   93,  15,  15,   4,   1, }, // 14 - R, P, Y
  {   93,  14,  14,   4,   1, }, // 15 - Thr, Mode

  {   93,  16,  16,   5,   3, }, // 16 - Tri
  {   93,  16,  16,   5,   3, }, // 17 - 
  
  {   93,  18,  18,   5,   3, }, // 18 - Quad+
  {   93,  18,  18,   5,   3, }, // 19

  {   93,  20,  20,   5,   3, }, // 20 - QuadX
  {   93,  20,  20,   5,   3, }, // 21

  {   93,  22,  22,   5,   3, }, // 22 - Bi
  {   93,  22,  22,   5,   3, }, // 23

  {   93,  24,  24,   5,   3, }, // 24 - Gimbal
  {   93,  24,  24,   5,   3, }, // 25

  {   93,  26,  26,   5,   3, }, // 26 - Y6
  {   93,  26,  26,   5,   3, }, // 27

  {   93,  28,  28,   5,   3, }, // 28 - Hex
  {   93,  28,  28,   5,   3, }, // 29

  {   93,  13,  11,   3,   1, }, // 30 - Altitde, Heading

  {   85, UP0,   1,   0,   0, }, // 31 - Acc, Bar, Mag

  {   34,   5,  33,   0,   0, }, // 32 - Profile
  {   86,  32, DN0,   0,   0, }, // 33 - Version

  {   37,  36,  35,  32,   0, }, // 34 - Load Profile
  {   38,  34,  36,  32,   0, }, // 35 - Save Profile
  {   39,  35,  34,  32,   0, }, // 36 - Clear Profile

  {   90,  87,  88,  34,   0, }, // 37 - Load Profile
  {   91,  87,  88,  35,   0, }, // 38 - Save Profile
  {   92,  87,  88,  36,   0, }, // 39 - Clear Profile

  {   93,  40,  40,   5,   3, }, // 40 - Flying Wing
  {   93,  40,  40,   5,   3, }, // 41

#ifdef LOGGING
  {   43,  33,   0,   0,   0, }, // 42 - Logging
  {   94,  45,  44,  42,   0, }, // 43 - Start Logging   #94
  {   95,  43,  45,  42,   0, }, // 44 - Delete LogFile  #95
  {   96,  44,  43,  42,   0, }, // 45 - Delay time      #96
#endif
  // Right Up  Down Left  Status      Status Flag: 1=Refresh Data, 2=2-Line Mode, 4-special case to refresh AutoLevel 
};

// configuration-dependent substates of state 'ESC out'
PROGMEM prog_char nextOutState[NUM_CONFIGS] = { 5, 16, 18, 20, 22, 24, 26, 28, 40, };

// charset used in profile names
static char      charsInName[] = " .01234567890_ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static char      profileMsg [] = " #00: 01234567  ";
static char      lineBuffer [17];

// ----------------

static uint8_t   ptr;
static uint8_t   buffer[100];

// ----------------

static int8_t    state                ;
static int8_t    profile              ;
static boolean   graphMode            ; // graphical output of servo signals
static boolean   charsInitialized     ; // chars for servo signals are initialized
static int8_t    logMode              ; // logging mode: -1 not available, 0 - disabled, 1 - enabled
static uint16_t  lineNo               ; // current record number in logFile
static uint8_t   logDelay             = 10; // x10 in milliseconds
static uint8_t   vers                 = 0;  // MultiWii Software Version 1.5
static uint8_t   backlight            = 100; // backlight intensity (0..255)
static uint8_t   minMaxMode           ;

// MWC data
#ifndef MINMAX
static int16_t   accSmooth   [3]      ;  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3]      ;  // from WMP+
static int16_t   altitude             ;  // from barometer
static int16_t   heading              ;  // from compas
static int16_t   servo       [4]      ; //= { 1500, 1500, 1500, 1500, };
static int16_t   motor       [6]      ; //= { 1500, 1500, 1500, 1500, 1500, 1500, }; // depends on configuration 
static int16_t   rcHysteresis[5]      ; //= { 1500, 1500, 1500, 1500, 1500,  };      // receiver input
static int16_t   angle       [2]      ;  // absolute angle inclination in deg (when Nunchuk/Acc present)
static uint8_t   nunchukPresent       ;
static uint8_t   i2C_AccPresent       ;
static uint8_t   i2C_BarometerPresent ;
static uint8_t   i2C_MagnetoPresent   ;
static uint8_t   levelModeParam       ;  // if level mode is a activated on the radio : channel 5 value superior to 1700
static uint16_t  cycleTime            ;  // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static int8_t    config               ;  // Tri, Quad, Hex, etc.
static uint8_t   P8[3], I8[3], D8[3]  ; // 8 bits is much faster and the code is much shorter
static uint8_t   rcRate8              ; // dual rate
static uint8_t   rcExpo8              ; // expo
static uint8_t   accStrength8         ; // AutoLevel Strength 0-250
static uint8_t   rollPitchRate        ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate              ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID            ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%
#else
static int16_t   accSmooth   [3][3]   ; //= { {0, 32767, -32767}, {0, 32767, -32767}, {0, 32767, -32767}, };  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3][3]   ; //= { {0, 32767, -32767}, {0, 32767, -32767}, {0, 32767, -32767}, };  // from WMP+
static int16_t   altitude       [3]   ; //=   {0, 32767, -32767};  // from barometer
static int16_t   heading        [3]   ; //=   {0, 32767, -32767};  // from compas
static int16_t   servo       [4][3]   ; //= { {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, };
static int16_t   motor       [6][3]   ; //= { {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, }; // depends on configuration 
static int16_t   rcHysteresis[5][3]   ; //= { {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767}, {1500, 32767, -32767} }; // receiver input
static int16_t   angle       [2][3]   ; //= { {0, 32767, -32767}, {0, 32767, -32767}, };  // absolute angle inclination in deg (when Nunchuk/Acc present)
static uint8_t   nunchukPresent       ;
static uint8_t   i2C_AccPresent       ;
static uint8_t   i2C_BarometerPresent ;
static uint8_t   i2C_MagnetoPresent   ;
static uint8_t   levelModeParam       ;  // if level mode is a activated on the radio : channel 5 value superior to 1700
static uint16_t  cycleTime            ;  // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static int8_t    config               ;  // Tri, Quad, Hex, etc.
static uint8_t   P8[3], I8[3], D8[3]  ; // 8 bits is much faster and the code is much shorter
static uint8_t   rcRate8              ; // dual rate
static uint8_t   rcExpo8              ; // expo
static uint8_t   accStrength8         ; // AutoLevel Strength 0-250
static uint8_t   rollPitchRate        ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate              ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID            ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%
#endif

// ----------------

void setup() {
  // make sure that the default chip select pin is set to
  // output, even if you don't use it:
  pinMode(10, OUTPUT);
  
#ifndef SRLCD
  lcd.begin(16, 2);   //no begin in ShiftRegLCD...
#endif
  lcd.clear();

  backlight = 255 - EEPROM.read(5);
  setBacklight();
  
  // init variables
  int16_t* aptr = (int16_t*) accSmooth;
  for (int8_t i = 0; i < 25; i++) {
    *aptr++ =  (i >= 8 && i < 23)? 1500 : 0;
#ifdef MINMAX 
    *aptr++ =  32767; // min
    *aptr++ = -32767; // max
#endif
  }
  
#ifdef TEST
  lcdPrint(12, 1, "Test");
#endif

#ifdef STATUS_LEDS
  pinMode(RED_LED,    OUTPUT);
  pinMode(YELLOW_LED, OUTPUT);
  pinMode(GREEN_LED,  OUTPUT);
  uint8_t bits = 1;
#endif

  for (int8_t i = 0; i < 32; i++) {
    lcdPrintProg(0, 0, intro + i);
#if defined(TEST) & defined(STATUS_LEDS)
    red   (bits & 4);
    yellow(bits & 2);
    green (bits & 1);
    
    bits *= 2;
    if (bits == 64) {
      bits = 1;
    }
#endif
    delay((i == 16)? 750 : 50);
  }
  yellow(false);
  
  if (!restoreParams()) {
    enterVersion();
  }
  setBacklight();

//  formatSigned(0, 0, logMode, 2, 0);
//  blinkLed(GREEN_LED, 250, 10);

  Serial.begin(SERIAL_COM_SPEED);

  boolean wait = true;
#ifdef LOGGING
  // see if the card is present and can be initialized:
  if (!SD.begin(CHIP_SELECT)) {
    setBacklight();
    lcdPrintProg(0, 1, noCard);
    logMode = -1;
    blinkLed(RED_LED, 500, 3);
  }
  setBacklight();
  if (logMode == 0) {
    wait = !playbackLogFile();
  }
#endif

  if (logMode != 1) {
#ifdef WAIT_FOR_INIT
    if (wait) {
      lcdPrintProg(0, 0, waitMsg1);    // wait for initialization
      lcdPrintProg(0, 1, waitMsg2);
  
      lcd.setCursor(15, 1);
      lcd.blink();
      enterKey();
      lcd.noBlink();
    }
#endif

    lcdPrintProg(0, 0, msgs[state]);
    lcdPrintProg(0, 1, reading);
  
    int readBytes = serialRead(true);
    lcdPrintProg(0, 1, readMsg);
    format(6, 1, readBytes, 2, 0);

    blinkLed(YELLOW_LED, 500);
    showStatusLine();
  }
#ifdef LOGGING
  else {
#ifdef WAIT_FOR_INIT
    lcdPrintProg(0, 0, waitMsg1);  // wait for initialization
    lcdPrintProg(0, 1, emptyLine);

  #ifndef TEST
    blinkLed(YELLOW_LED, 500, 10); // wait 20 seconds before start logging to give the MWC board the chance to initialize
  #endif
#endif

    openLog();
  }
#endif
}

void loop() {
#ifdef LOGGING
  if (logMode == 1) {
    doLog();
    return;
  }
#endif  

  uint8_t statusFlags = getNextState(state, STATUS);
  if (statusFlags & 1) {
    // update the current values from MultiWii board and refresh display
    serialRead(false);
    if (state != 0) {
      showStatusLine();
    }
    else if (nunchukPresent || i2C_AccPresent) {
      // refresh Autolevel state
      lcd.setCursor(15, 1);
      lcd.write('0' + levelModeParam);
    }
  }

#ifdef TEST
  incrementValues();
#endif      

  unsigned long startTime = millis();
  int8_t        key       = getKeyPressed();
  unsigned long endTime   = millis();
  if (key >=0) {
    int8_t next = getNextState(state, key);
#ifdef LOGGING
    if (next == 42 && logMode == -1) {
      next = (key == KEY_UP)? 33 : 0; // disable state 42 "Logging" when initialisation of SD Card failed 
    }
#endif
    
#ifdef TEST
    if (key == KEY_SELECT && state == 5) {
      config = (config + 1) % NUM_CONFIGS;
    }
#endif      

    if (next >= 80) { // handle pseudo-states
      switch (next) {
        case 80:                                 // PID roll
        case 81:                                 // PID pitch
        case 82: { enterPid  (next-80); break; } // PID yaw

        case 83:                                 // dual rate, expo, autolevel strength
        case 84: { enterRates(next-83); break; } // roll/pitch rate, yaw rate, throttle pid

        case 85: { resetValues();  state = 0; break; } // reset to default
        case 86: { enterVersion(); enterBacklight(); state = 33; break; } // select Version of MWC software and backlight intensity

        case 87: { profile = (profile+15) % 16; break; } // decrement profile
        case 88: { profile = (profile+ 1) % 16; break; } // increment profile

        case 90: { state = restoreProfile()? 34:37; break; } // load profile
        case 91: { storeProfile(); state = 35;      break; } // save profile
        case 92: { deleteProfile();                 break; } // clear profile

        case 93: {
          if (key == KEY_SELECT || endTime - startTime > 500) {
#ifdef GRAPHIC
            graphMode = !graphMode;
            showGraphicMode();
#endif
          }
          else {
#ifdef MINMAX
            minMaxMode = (minMaxMode + 1) % 3; 
            showMinMax();
#endif
          }
          break; 
        }

#ifdef LOGGING
        case 94: {                                      // start logging
          boolean flag = confirm(haltReqMsg);
          if (flag) {
            logMode = 1;
            storeParams();
            setError(logEnbMsg);
          }
          openLog();
          break; 
        }
        case 95: { deleteLogFile(); break; }            // delete logfile
        case 96: { enterLogDelay(); break; }            // delay time
#endif

        case 99: { state = pgm_read_byte_near(nextOutState + getConfig()); break; } // select state depending on config
      }
      next = state;
    }
    
    if (next >= 0 && logMode < 1) {
      state = next;
      lcdPrintProg(0, 0, msgs[state]);
      prog_char* line2 = (getNextState(state, STATUS) & 2)? msgs[state+1] : emptyLine;
      lcdPrintProg(0, 1, line2);
      showStatusLine();
    }
  }
}

int8_t getNextState(int8_t state, int8_t key) { 
  int8_t col = (key == KEY_SELECT)? KEY_RIGHT : (key == STATUS)? STATUS-1 : key;
#ifdef MINMAX
  if (state >= 3 && state <= 5 && col == KEY_RIGHT) {
    showMinMax();
  }
#endif  
  return pgm_read_byte_near(nextState[state] + col); 
}

int8_t getConfig() { 
  return (config < 0 || config >= NUM_CONFIGS)? 0 : config; 
}

#ifdef TEST
void incrementValues() {
  #ifndef MINMAX
  for (int8_t i = 0; i < 5; i++) {rcHysteresis[i] = (rcHysteresis[i] - 1000 + 20) % 1000 + 1000;} // receiver input
  for (int8_t i = 0; i < 2; i++) {angle       [i] = (angle[i] + 180 + 2 + 3 * i)  % 360 - 180;}
  #else
  for (int8_t i = 0; i < 5; i++) {
    rcHysteresis[i][0] =    (rcHysteresis[i][0] - 1000 + 20) % 1000 + 1000; // receiver input
    rcHysteresis[i][1] = min(rcHysteresis[i][0], rcHysteresis[i][1]);
    rcHysteresis[i][2] = max(rcHysteresis[i][0], rcHysteresis[i][2]);
  }
  for (int8_t i = 0; i < 2; i++) {
    angle[i][0] =    (angle[i][0] + 180 + 2 + 3 * i) % 360 - 180;
    angle[i][1] = min(angle[i][0], angle[i][1]);
    angle[i][2] = max(angle[i][0], angle[i][2]);
  }
  #endif      
}
#endif

// ----------------

void setError(prog_char const msg[]) {
  lcdPrintProg(0, 0, msg);
  uint8_t offs = 0;
  while (true) {
    lcdPrintProg(0, 1, haltMsg1 + offs);
    offs = 17 - offs;
    blinkLed(RED_LED, 2000);
  }
}

void blinkLed(uint8_t led, uint16_t del) {
#ifdef STATUS_LEDS
  del /= 2;
  digitalWrite(led, HIGH); 
  delay(del); 
  digitalWrite(led, LOW); 
#endif
  delay(del);    
}

void blinkLed(uint8_t led, uint16_t del, int8_t repeat) {
  for (int8_t i = 0; i < repeat; i++) {blinkLed(led, del);}
}

void resetValues() {
  if (confirm(resetMsg)) {
    P8[0] = 40; I8[0] = 30; D8[0] = 13;
    P8[1] = 40; I8[1] = 30; D8[1] = 13;
    P8[2] = 80; I8[2] =  0; D8[2] =  0;
    
    rcRate8       = 90;
    rcExpo8       = 65;
    accStrength8  = 100;
    rollPitchRate = 0;
    yawRate       = 0;
    dynThrPID     = 0;
    
    serialWrite();
  }
}

// ----------------

void lcdPrint(int8_t col, int8_t row, char const msg[]) {
  lcd.setCursor(col, row);
  lcd.print(msg);
}

void lcdPrintProg(int8_t col, int8_t row, prog_char const msg[]) {
  char* lptr = lineBuffer;
  char last  = -1;
  for (int8_t i = 0; last != '\0' && i < 16; i++) {
    last    = pgm_read_byte_near(msg + i);
    *lptr++ = last;
  }
  *lptr++ = '\0';

  lcd.setCursor(col, row);
  lcd.print(lineBuffer);
}

// ----------------

void lcdWrite(int8_t col, int8_t digit) {
  lcd.setCursor(col, 1);
  lcd.write((digit)? '1' : '0');
}

void showStatusLine() {
  switch (state) {
    case 0: {
      int8_t tconf = getConfig();
      lcdPrintProg(0, 1, configs[tconf]);

      lcdPrintProg(7, 1, levelMsg);
      if (i2C_AccPresent || nunchukPresent) {
        lcdWrite(15, levelModeParam);
      }
      break;
    }
    case 31: {
      lcdPrintProg( 0, 1, statusMsg);
      if (i2C_AccPresent) {
        lcdPrintProg( 0, 1, accMsg);
      }
      lcdWrite( 3, i2C_AccPresent || nunchukPresent);
      lcdWrite( 9, i2C_BarometerPresent            );
      lcdWrite(14, i2C_MagnetoPresent              );
      break;
    }
    
    case 6:   // PID Roll
    case 7:   // PID Pitch
    case 8: { // PID Yaw
      format      ( 2, 1,  P8[state-6], 4, 2);    
      format      ( 7, 1,  I8[state-6], 5, 4);    
      formatSigned(13, 1, -D8[state-6], 3, 0);    
      break;
    }
    
    case 9: { // Dual Rate, Expo, AutoLevel Strength
      format( 0, 1, rcRate8,      4, 0);
      format( 6, 1, rcExpo8,      4, 0);
      format(12, 1, accStrength8, 4, 0);
      break;
    }
    case 10: { // R/P Rate, Yaw Rate, Throttle PID
      format( 0, 1, rollPitchRate, 4, 0);
      format( 6, 1, yawRate,       4, 0);
      format(12, 1, dynThrPID,     4, 0);
      break;
    }
    
    case 11: { // Gyro
      for (uint8_t i = 0; i < 3; i++) { formatSigned(i*6, 1, gyroData[i], 4, 0); }
      break;
    }
    case 12: { // Acc
      for (uint8_t i = 0; i < 3; i++) { formatSigned(i*6, 1, accSmooth[i], 4, 0); }
      break;
    }
    case 13: { // Angles, Cycle
      if (!graphMode) {
        formatSigned( 0, 1, angle [0], 4, 0);
        formatSigned( 6, 1, angle [1], 4, 0);
      }
#ifdef GRAPHIC
      else {
        drawAngle(0);
        drawAngle(1);
      }
#endif
      format(12, 1, cycleTime, 4, 0);
      break;
    }

    case 14: { // Roll, Pitch, Yaw
      formatScaledSym( 0, 1, rcHysteresis[ROLL    ]); 
      formatScaledSym( 6, 1, rcHysteresis[PITCH   ]);
      formatScaledSym(12, 1, rcHysteresis[YAW     ]);
      break;
    }
    case 15: { // Throttle, Mode
      formatScaledAbs( 6, 1, rcHysteresis[THROTTLE]);
      formatScaledSym(12, 1, rcHysteresis[AUX1    ]);
      break;
    }

    case 16: { // Left, Right, Rudder, Rear
      formatScaledAbs( 3, 0, motor[2]);
      formatScaledAbs(12, 0, motor[1]);

      formatScaledSym( 3, 1, servo[0]);
      formatScaledAbs(12, 1, motor[0]);
      break;
    }

    case 18: { // Front, Rear, Left, Right
      formatScaledAbs( 3, 0, motor[3]);
      formatScaledAbs(12, 0, motor[0]);

      formatScaledAbs( 3, 1, motor[2]);
      formatScaledAbs(12, 1, motor[1]);
      break;
    }

    case 20: { // Front Left, Front Right, Rear Left, Rear Right
      formatScaledAbs( 3, 0, motor[3]);
      formatScaledAbs(12, 0, motor[1]);

      formatScaledAbs( 3, 1, motor[2]);
      formatScaledAbs(12, 1, motor[0]);
      break;
    }

    case 22: { // Motor 0, Motor 1, Servo 0, Servo 1 
      formatScaledAbs( 3, 0, motor[0]);
      formatScaledAbs(12, 0, motor[1]);

      formatScaledSym( 3, 1, servo[0]);
      formatScaledSym(12, 1, servo[1]);
      break;
    }

    case 24: { // Roll, Pitch
      formatScaledSym( 7, 0, servo[1]);
      formatScaledSym( 7, 1, servo[0]);
      break;
    }

    case 26: { // Up (Left, Mid/Rear, Right), Down (Left, Mid, Right)
      formatScaledAbs( 2, 0, motor[2]);
      formatScaledAbs( 7, 0, motor[0]);
      formatScaledAbs(12, 0, motor[1]);

      formatScaledAbs( 2, 1, motor[5]);
      formatScaledAbs( 7, 1, motor[3]);
      formatScaledAbs(12, 1, motor[4]);
      break;
    }

    case 28: { // Front (Left, Mid, Right), Rear (Left, Mid, Right)
      formatScaledAbs( 2, 0, motor[3]);
      formatScaledAbs( 7, 0, motor[4]);
      formatScaledAbs(12, 0, motor[1]);

      formatScaledAbs( 2, 1, motor[2]);
      formatScaledAbs( 7, 1, motor[5]);
      formatScaledAbs(12, 1, motor[0]);
      break;
    }

    case 30: { // Alt, Heading
      formatSigned( 0, 1, altitude, 4, 0);
      formatSigned( 6, 1, heading , 4, 0);
      break;
    }
    
    case 37:
    case 38:
    case 39: {
      profileMsg[2] = ('0' + profile/10);
      profileMsg[3] = ('0' + profile%10);
      getProfileName(profileMsg+6);
      
      lcdPrint(0, 1, profileMsg);
      break;
    }

    case 40: { // Servo 0-3 
      formatScaledAbs( 3, 0, servo[0]);
      formatScaledAbs(12, 0, servo[1]);

      formatScaledSym( 3, 1, servo[2]);
      formatScaledSym(12, 1, servo[3]);
      break;
    }
  }
}

// ----------------

void enterPid(int8_t index) {
  int val[3], minv[3] = { 0, 0, -40 }, maxv[3] = { 200, 250, 0 };

  val[0] = P8[index]; 
  val[1] = I8[index]; 
  val[2] = -D8[index];
  
  int8_t local   = 0; // local state (-1..4)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;

  yellow(true);
  while (local >= 0 && local < 3) {
    boolean update = false;

    int8_t key = getKey();
    if (key < 0) {
      lastKey = key;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_LEFT  : { 
          local--;
          update = true; 
          break; 
        }
        case KEY_RIGHT : { 
          local++; 
          update = true; 
          if (local < 3) { // handle RIGHT in third column as SELECT
            break;
          }
        }
        case KEY_SELECT: { 
          lcdPrintProg(0, 1, writing);
          
          P8[index] =  val[0];
          I8[index] =  val[1];
          D8[index] = -val[2];
          
          serialWrite();
          local = -1;
          break; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_UP) {
              val[local] = min(val[local]+1, maxv[local]);
            }
            else {
              val[local] = max(val[local]-1, minv[local]);
            }
            switch (local) {
              case 0: { format      ( 2, 1, val[0], 4, 2); break; }
              case 1: { format      ( 7, 1, val[1], 5, 4); break; }    
              case 2: { formatSigned(13, 1, val[2], 3, 0); break; } 
            }
            repeat++;
          }
          break; 
        }
      }
    }
    
    if (update || (count % 20) == 0) {
      char c[3] = {' ', ' ', ' '};
      if ((update || (count / 20) % 2 == 0) && (local >= 0 && local < 3)) {
        c[local] = '\176';
      }
      for (int i = 0; i < 3; i++) {
        lcd.setCursor(max(1, i * 6), 1);
        lcd.write(c[i]);
      }
    }
    count++;
    delay(50);
  }
  waitForKeyUp();
  yellow(false);
}

void enterRates(int8_t index) {
  static char leftCursor2 [] = "\177 ";
  static char rightCursor2[] = " \176";
  static char blank2      [] = "  ";
  
  uint8_t val[3];
 
  if (index == 0) {
    val[0] = rcRate8;
    val[1] = rcExpo8;
    val[2] = accStrength8;
  }
  else {
    val[0] = rollPitchRate;
    val[1] = yawRate;
    val[2] = dynThrPID;
  }
  
  int8_t local   = 0;  // local state (-1..4)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;

  yellow(true);
  while (local >= 0 && local < 3) {
    boolean update = false;
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_LEFT  : { 
          local--;
          update = true;
          break; 
        }
        case KEY_RIGHT : { 
          local++; 
          update = true;
          if (local < 3) { // handle RIGHT in third column as SELECT
            break;
          }
        }
        case KEY_SELECT: { 
          lcdPrintProg(0, 1, writing);
          
          if (index == 0) {
            rcRate8       = val[0];
            rcExpo8       = val[1];
            accStrength8  = val[2];
          }
          else {
            rollPitchRate = val[0];
            yawRate       = val[1];
            dynThrPID     = val[2];
          }
          
          serialWrite();
          local = -1;
          break; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_UP) {
              val[local] = min(val[local]+1, (index == 0 && local == 2)? 250 : 100);
            }
            else {
              val[local] = max(val[local]-1, 0);
            }
            format(local * 6, 1, val[local], 4, 0);
            repeat++;
          }
          break; 
        }
      }
    }
    
    if (update || (count % 20) == 0) {
      char* left  = blank2;
      char* right = blank2;
      
      if (update || (count / 20) % 2 == 0) {    
        switch (local) {
          case 0: { left = leftCursor2;                        break; }
          case 1: { left = rightCursor2, right = leftCursor2;  break; }
          case 2: {                      right = rightCursor2; break; } 
        }
      }
      lcdPrint( 4, 1, left);
      lcdPrint(10, 1, right);
    }
    count++;
    delay(50);
  }
  waitForKeyUp();
  yellow(false);
}

boolean enterName(char name[]) {
  int8_t local   = 0; // local state (-1..8)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;
  
  for (int8_t i = 0; i < 8; i++) {
    if (name[i] == '-') {
      name[i] = ' ';
    }
  }
  
  int8_t idx = findChar(name[local]);
  
  lcd.cursor();
  lcd.setCursor(local + 6, 1);
  lcd.print(name[local]);
  lcd.setCursor(local + 6, 1);
  yellow(true);
  
  while (local >= 0 && local < 8) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      boolean refresh = false;
      switch (key) {
        case KEY_LEFT: { 
          local--;
          refresh = (local >= 0);
          if (refresh) {
            idx = findChar(name[local]);
          }
          break; 
        }
        case KEY_RIGHT: { 
          local++; 
          refresh = (local < 8);
          if (refresh) { // handle RIGHT in third column as SELECT
            idx = findChar(name[local]);
          }
          break;
        }
        case KEY_SELECT: {
          local = 8;
          break; 
        }
        case KEY_DOWN: case KEY_UP: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_DOWN && idx < sizeof(charsInName)-2) {
              idx++;
              refresh = true;
            }
            else if (key == KEY_UP && idx > 0) {
              idx--;
              refresh = true;
            }
            if (refresh) {
              name[local] = charsInName[idx];
            }
          }
          break; 
        }
      }
      if (refresh) {
        lcd.setCursor(local + 6, 1);
        lcd.print(name[local]);
        lcd.setCursor(local + 6, 1);
        repeat++;
      }
    }
    count++;
    delay(50);
  }
  waitForKeyUp();
  yellow(false);
  lcd.noCursor();
  
  return (local == 8);
}

int findChar(char ch) {
  int res = -1;
  for (int i = 0; res < 0 && i < sizeof(charsInName)-1; i++) {
    if (charsInName[i] == ch) {
      res = i;
    }
  }
  if (res < 0) {
    res = 0;
  }
  return res;
}

void enterVersion() {
  yellow(true);
  lcdPrintProg(0, 0, versMsg);

  uint8_t old = max(vers, 0);
  int8_t  key = -1;
  while (key < 0 || key == KEY_UP || key == KEY_DOWN) {
    if (key == KEY_UP) {
      vers = max(vers-1, 0);
    }
    else if (key == KEY_DOWN) {
      vers = min(vers+1, sizeof(frameLength)-1);
    }
    lcdPrintProg(4, 1, versions[vers]);
    delay(50);
    key = getKeyPressed();
  }
  if (key != KEY_LEFT) {
    storeParams();
  }
  else {
    vers = old;
  }
  yellow(false);
}

void enterBacklight() {
#ifdef BACKLIGHT
  yellow(true);
  lcdPrintProg(0, 0, bLightMsg);
  lcdPrintProg(0, 1, emptyLine);

  uint8_t val = backlight;
  format(7, 1, val, 3, 0);

  int     repeat  = 0;
  int     count   = 0;
  int8_t  lastKey = -1;

  yellow(true);
  while (true) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_RIGHT: case KEY_SELECT: {
          backlight = val;
          storeParams();
          // intended fall through to KEY_LEFT
        }
        case KEY_LEFT: { 
          waitForKeyUp();
          setBacklight();
          yellow(false);
          return; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            val = (key == KEY_UP)? min(val+5, 250) : max(val-5, 0);
            analogWrite(10, val);
            format(7, 1, val, 3, 0);
            repeat++;
          }
          break; 
        }
      }
    }
    count++;
    delay(50);
  }
#endif
}

#ifdef LOGGING
void enterLogDelay() {
  lcdPrint(10, 0, ":");
  format  ( 5, 1, logDelay, 3, 0);
  lcdPrint( 8, 1, "0 ms");
  
  int     val     = logDelay;
  int     repeat  = 0;
  int     count   = 0;
  int8_t  lastKey = -1;

  yellow(true);
  while (true) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_RIGHT: case KEY_SELECT: {
          logDelay = val;
          storeParams();
          // intended fall through to KEY_LEFT
        }
        case KEY_LEFT: { 
          waitForKeyUp();
          yellow(false);
          return; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            val = (key == KEY_UP)? min(val+5, 250) : max(val-5, 0);
            format(5, 1, val, 3, 0);
            repeat++;
          }
          break; 
        }
      }
    }
    count++;
    delay(50);
  }
}
#endif

// ----------------

void format(int8_t col, int8_t row, int val, int8_t len, int8_t prec) {
  formatNum(col, row, val, len, prec, false);
}

void formatSigned(int8_t col, int8_t row, int val, int8_t len, int8_t prec) {
  formatNum(col, row, val, len, prec, true);
}

void formatNum(int8_t col, int8_t row, int val, int8_t len, int8_t prec, bool sgn) {
  char temp[len+1];
  int8_t low = 0;
  if (sgn) {
    temp[0] = (val > 0)? '+' : (val < 0)? '-' : ' ';
    low++;
  }
  val = abs(val);
  for (int8_t i = len-1; i >= low; i--) {
    if (len-i == prec) {
      temp[i] = '.';
    }
    else {
      temp[i] = '0' + (val % 10);
      val    /= 10;
    }
  }
  if (val > 0) { // overflow
    for (int8_t i = len-1; i >= 0; i--) {
      temp[i] = '#';
    }
  }
  temp[len] = '\0';
  lcdPrint(col, row, temp);
}

//int scaleServo   (int val) {return (val - 1500)/5; }
//int scaleServoUns(int val) {return (val - 1000)/10;}

void formatScaledSym(int8_t col, int8_t row, int val) { // -100..+100
  val = (val - 1500)/5;
  if (!graphMode) {
    formatSigned(col, row, val, 4, 0);
  }
#ifdef GRAPHIC
  else {
    formatGraphic(col, row, (val+100)/2);
  }
#endif
}

void formatScaledAbs(int8_t col, int8_t row, int val) { // 0..100
  val = (val - 1000)/10;
  if (!graphMode) {
    formatSigned(col, row, val, 4, 0);
  }
#ifdef GRAPHIC
  else {
    formatGraphic(col, row, val);
  }
#endif
}

#ifdef GRAPHIC
void formatGraphic(int8_t col, int8_t row, int val) { // 0..100
  if (!charsInitialized) {
    charsInitialized = true;
  
    static byte bars   [8] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, };
    static byte lesser [8] = { 0x00, 0x04, 0x0C, 0x1C, 0x0C, 0x04, 0x00, 0x15, };
    static byte greater[8] = { 0x00, 0x04, 0x06, 0x07, 0x06, 0x04, 0x00, 0x15, };

    byte pattern = 0x10;
    for (int8_t i = 0; i <= 5; i++) {
      for (int8_t j = 0; j < 7; j++) {
        bars[j] = pattern;
      }
      lcd.createChar(i, bars);
      pattern >>= 1;
    }
    lcd.createChar(6, lesser);
    lcd.createChar(7, greater);
  }  
  
  char bar[] = { 5, 5, 5, 5 };
  
  if      (val <    0) { bar[0]      = 6; }           // <...
  else if (val >= 100) { bar[3]      = 7; }           // ...>
  else                 { bar[val/25] = (val%25)/5; }  // ..|.
  lcd.setCursor(col, row);
  lcd.write(bar[0]); 
  lcd.write(bar[1]); 
  lcd.write(bar[2]); 
  lcd.write(bar[3]); 
}

void showGraphicMode() {
  lcdPrintProg( 0, 0, emptyLine);  
  lcdPrintProg( 0, 1, emptyLine);  
  lcdPrintProg( 1, 0, graphicMsg);
  lcdPrintProg(10, 0, onOffMsg[graphMode? 1:0]);
  blinkLed(YELLOW_LED, 750);
}

/* ----- */

static uint8_t angleChars[4][8];

void drawAngle(int8_t index) {
  // clear char arrays
  uint8_t *cptr = angleChars[0];
  for (int8_t i = 0; i < 32; i++) {
    *cptr++ = 0;
  }

#ifndef MINMAX
  int16_t alpha = -angle[index];
#else
  int16_t alpha = -angle[index][minMaxMode];
#endif
  float   ang   = radians(alpha);
  float   sinA  = sin(ang);
  float   cosA  = cos(ang);
  
  if (abs(sinA) > 0.25f) {
    angleChars[0][4] = 0x14;  // mark horizon
    angleChars[3][4] = 0x05;
  }
  
  int8_t x0 = (int8_t) (- cosA * 10.0f + 0.5f);
  int8_t y0 = (int8_t) (- sinA * 10.0f + 0.5f);
  
  int8_t charOffset = index * 4;
  drawLine(charOffset, x0, y0, -x0, -y0);
  
  int8_t x1 = (int8_t) (- sinA * 3.0f + 0.5f);
  int8_t y1 = (int8_t) (  cosA * 3.0f + 0.5f);

  drawLine(charOffset, 0, 0, x1, y1);

  for (int8_t i = 0; i < 4; i++) {
    lcd.createChar(charOffset + i, angleChars[i]);
    lcd.setCursor (index * 6  + i, 1);
    lcd.write     (charOffset + i);
  }
  
  charsInitialized = false;  
}

void drawLine(int8_t charOffset, int8_t x0, int8_t y0, int8_t x1, int8_t y1) {
  // Bresenham again...
  
  int8_t dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int8_t dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int8_t err = dx+dy, e2; /* error value e_xy */
 
  for(;;){  /* loop */
    setPixel(x0, y0);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
    if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
  }  
}

void setPixel(int8_t x, int8_t y) { 
  // x: -10..+9, y: -3..+4
  
  if (x >= -10 && x < 10 && y >= -3 && y < 5) {
    int8_t nrow = 4 - y;   // 0..7
    int8_t ncol = x + 10;  // 0..19
    int8_t idx  = ncol / 5;
    int8_t pcol = ncol % 5;
    
    angleChars[idx][nrow] |= (0x10 >> pcol);
  }
}
#endif // GRAPHIC

// ----------------

#ifdef MINMAX
void format         (int8_t col, int8_t row, int arr[], int8_t len, int8_t prec) {format      (col, row, arr[minMaxMode], len, prec);}
void formatSigned   (int8_t col, int8_t row, int arr[], int8_t len, int8_t prec) {formatSigned(col, row, arr[minMaxMode], len, prec);}
void formatScaledSym(int8_t col, int8_t row, int arr[]) {formatScaledSym(col, row, arr[minMaxMode]);}
void formatScaledAbs(int8_t col, int8_t row, int arr[]) {formatScaledAbs(col, row, arr[minMaxMode]);}

  #ifdef GRAPHIC
void formatGraphic  (int8_t col, int8_t row, int arr[]) {formatGraphic  (col, row, arr[minMaxMode]);}
  #endif

void showMinMax() {
  lcdPrintProg(0, 0, emptyLine);  
  lcdPrintProg(0, 1, emptyLine);  
  lcdPrintProg(1, 0, minMaxMsg);
  lcdPrintProg(7, 0, minMaxDsc[minMaxMode]);
  blinkLed(YELLOW_LED, 750);
}
#endif

// ----------------

boolean confirm(prog_char const msg[]) {
  yellow(true);
  lcdPrintProg(0, 0, msg);
  lcdPrintProg(0, 1, yesNoMsg);
    
  int8_t key = -1;
  while (key != KEY_LEFT && key != KEY_RIGHT) {
    delay(50);
    key = getKeyPressed();
  }      

  yellow(false);
  return (key == KEY_RIGHT);
}

int8_t enterKey() {
  yellow(true);
  int8_t key = -1;
  while ((key = getKeyPressed()) < 0);
  yellow(false);
  return key;
}

int8_t getKeyPressed() {
  int8_t key = getKey();
  if (key >= 0) {
    waitForKeyUp();
  }
  return key;
} 

void waitForKeyUp() {
  while (rawGetKey() >= 0);
}

int8_t getKey() {
  static int8_t oldKey = -1;

  int8_t key = rawGetKey();
  if (key != oldKey) {
    delay(50);		// wait for debounce time
    key = rawGetKey();
    if (key != oldKey) {			
      oldKey = key;
    }
  }
  return key;
}

#ifndef HWKEYS   //no hw keys!
int adc_key_val[NUM_KEYS] = { 30, 150, 360, 535, 760 };

// Convert ADC value to key number
int8_t rawGetKey() {
  uint16_t input  = analogRead(0);   // read the value from the sensor  
  int8_t   result = -1;
  for (uint8_t k = 0; result < 0 && k < NUM_KEYS; k++) {
    if (input < adc_key_val[k]) {
      result = k;
    }
  }
  return result;
}
#else        // hw keys used
int8_t rawGetKey() {
  int8_t result = -1;
  if      (digitalRead(HW_RIGHT) ==0) {result = KEY_RIGHT; }
  else if (digitalRead(HW_UP)    ==0) {result = KEY_UP;    }
  else if (digitalRead(HW_DOWN)  ==0) {result = KEY_DOWN;  }
  else if (digitalRead(HW_LEFT)  ==0) {result = KEY_LEFT;  }
  else if (digitalRead(HW_SELECT)==0) {result = KEY_SELECT;}
  return result;
}
#endif

// ----------------

//#define serialize16(a) buffer[ptr++] = a; buffer[ptr++] = a>>8&0xff
//#define serialize8(a)  buffer[ptr++] = a;
//#define deserialize16() ((int16_t) (buffer[ptr++] + 256 * buffer[ptr++]))
//#define deserialize8() buffer[ptr++]

void    serialize16(int16_t a) {buffer[ptr++] = a; buffer[ptr++] = a>>8&0xff;}
void    serialize8 (int8_t  a) {buffer[ptr++] = a;}

int16_t deserialize16() {uint8_t low = buffer[ptr++]; uint8_t hi = buffer[ptr++]; return (int16_t) (hi * 256 + low);}
int8_t  deserialize8()  {return buffer[ptr++];}

#ifndef TEST
int serialRead(boolean init) {
  static boolean ignoreMode = false;

  int readBytes = 0;
  if (ignoreMode) {
    blinkLed(RED_LED, 20);
    delay(100);
  }
  else {
    boolean frameErr  = false;
    uint8_t len       = frameLength[vers];
    while (true) {
      Serial.flush();
    
      boolean avail;
      int     retry = 0;
      do {
        Serial.write('A');
        delay(50);
        avail = Serial.available();
        if (!avail) {
          retry++;
          if (retry > 600) {
            setError(timeoutError);
          }
          else if (retry > 10) {
            if (init && getKey() == KEY_LEFT) {
              // cancel loop
              lcdPrintProg(0, 1, cancelRead);
              blinkLed(RED_LED, 250, 6);
              ignoreMode = true;
              logMode    = -1;
              return 0;
            }
            blinkLed(RED_LED, 4);
          }
        }
      }
      while (!avail); // 30 seconds until timeout
    
      blinkLed(YELLOW_LED, 4);
        
      char prefix = Serial.read();
      delay(2);
    
      if (prefix == 'A') {
        readBytes = 0;
        buffer[readBytes++] = prefix;
        while (readBytes <= len && Serial.available()) {
          buffer[readBytes++] = Serial.read();
          delay(2);
       }
        if (readBytes > len && buffer[readBytes-1] == 'A') {
          deserialize();
          break;
        }
        else if (!frameErr) {
          frameErr = true;
        }
        else {
          vers = -1;       // set version invalid to force version request after next start
          storeParams();
          setError(frameError);
        }
      }
    }
  }
  return readBytes;
}
#else
boolean myInit;

int serialRead(boolean init) {
  if (!myInit) {
    myInit = true;
    ptr = 0;
  
    serialize8('A');
    for (int8_t i=0; i<3; i++) serialize16(i * 100);
    for (int8_t i=0; i<3; i++) serialize16(i * 200);
    serialize16(0);
    if (vers >= 1) {
      serialize16(1);
    }
    int8_t numServos = (vers >= 2)? 4 : 2;  
    for (int8_t i=0; i<numServos; i++) serialize16(1000 + i * 100);

    for (int8_t i=0; i<6; i++) serialize16(1200 + i * 100);
    for (int8_t i=0; i<5; i++) serialize16(1500 + i * 100);      
    serialize8(1); 
    if (vers >= 1) {
      serialize8(1);
      serialize8(2);
      serialize8(3);
    }  
    serialize8(0);
    serialize16(6008);
    for (int8_t i=0; i<2; i++) serialize16(90);
    serialize8(1);  // config
    for (int8_t i=0; i<3; i++) {serialize8(7); serialize8(15); serialize8(31);}
    serialize8(11); 
    serialize8(22);
    serialize8(33);
    serialize8(44); 
    serialize8(55);
    serialize8(66);
    serialize8('A');
      
    deserialize();
  }
  else {
    ptr = frameLength[vers] + 1;
  }

  blinkLed(YELLOW_LED, 4);
  delay(100);

  return ptr;
}
#endif

#ifndef MINMAX
void deserialize() {
  ptr = 1;
  
  int8_t i;
  for (i=0; i<3; i++) {accSmooth[i]    = deserialize16();}
  for (i=0; i<3; i++) {gyroData [i]    = deserialize16();}
  altitude          = deserialize16();

  if (vers >= 1) {
    heading         = deserialize16();
  }

  int8_t numServos = (vers >= 2)? 4 : 2;  
  for (i=0; i<numServos; i++) {servo[i] = deserialize16();}

  for (i=0; i<6; i++) {motor       [i] = deserialize16();}
  for (i=0; i<5; i++) {rcHysteresis[i] = deserialize16();}

  nunchukPresent    = deserialize8();

  if (vers >= 1) {
    i2C_AccPresent       = deserialize8();
    i2C_BarometerPresent = deserialize8();
    i2C_MagnetoPresent   = deserialize8();
  }  

  levelModeParam    = deserialize8(); 
  cycleTime         = deserialize16();
  
  for (i=0; i<2; i++) {angle[i]        = deserialize16();}
  
  config            = deserialize8(); 

  for (i=0; i<3; i++) {
    P8[i]           = deserialize8(); 
    I8[i]           = deserialize8();
    D8[i]           = deserialize8();
  }

  rcRate8           = deserialize8();  
  rcExpo8           = deserialize8(); 
  accStrength8      = deserialize8(); 
  rollPitchRate     = deserialize8(); 
  yawRate           = deserialize8(); 
  dynThrPID         = deserialize8(); 
}
#else
void deserialize16(int arr[]) {
  arr[0] = deserialize16();
  arr[1] = min(arr[0], arr[1]);
  arr[2] = max(arr[0], arr[2]);
}

void deserialize() {
  ptr = 1;
  
  int8_t i;
  for (i=0; i<3; i++) {deserialize16(accSmooth[i]);}
  for (i=0; i<3; i++) {deserialize16(gyroData [i]);}
  deserialize16(altitude);

  if (vers >= 1) {
    deserialize16(heading);
  }
  for (i=0; i<2; i++) {deserialize16(servo       [i]);}
  for (i=0; i<6; i++) {deserialize16(motor       [i]);}
  for (i=0; i<5; i++) {deserialize16(rcHysteresis[i]);}

  nunchukPresent    = deserialize8();

  if (vers >= 1) {
    i2C_AccPresent       = deserialize8();
    i2C_BarometerPresent = deserialize8();
    i2C_MagnetoPresent   = deserialize8();
  }  

  levelModeParam    = deserialize8(); 
  cycleTime         = deserialize16();
  
  for (i=0; i<2; i++) {deserialize16(angle[i]);}
  
  config            = deserialize8(); 

  for (i=0; i<3; i++) {
    P8[i]           = deserialize8(); 
    I8[i]           = deserialize8();
    D8[i]           = deserialize8();
  }

  rcRate8           = deserialize8();  
  rcExpo8           = deserialize8(); 
  accStrength8      = deserialize8(); 
  rollPitchRate     = deserialize8(); 
  yawRate           = deserialize8(); 
  dynThrPID         = deserialize8(); 
}
#endif

void serialWrite() {
  ptr = 0;
  serialize8('C');
  for (int8_t i=0; i<3; i++) { serialize8(P8[i]); serialize8(I8[i]); serialize8(D8[i]); }
  serialize8(rcRate8      );
  serialize8(rcExpo8      );
  serialize8(accStrength8 );
  serialize8(rollPitchRate);
  serialize8(yawRate      );
  serialize8(dynThrPID    );
  
  Serial.write(buffer, ptr);

  lcdPrintProg(0, 1, writeMsg);
  format(7, 1, ptr, 2, 0);

  blinkLed(GREEN_LED, 500, 3); // signal "serial write"
  
  serialRead(false); // reload the frame
}

// ----------------

void storeProfile() {
  getProfileName(lineBuffer); 
  if (enterName(lineBuffer)) {
    storeProfile(lineBuffer);
  }
}

bool restoreProfile() {
  uint16_t adr     = constrain((profile+1), 1, 16) * 32 + 8; // skip name
  char     hdr     = EEPROM.read(adr++);
  boolean  succ    = (hdr == 'C');  

  if (succ) {
    for (int8_t i=0; i<3; i++) {
      P8[i]       = EEPROM.read(adr++); 
      I8[i]       = EEPROM.read(adr++); 
      D8[i]       = EEPROM.read(adr++);
    }
    rcRate8       = EEPROM.read(adr++);
    rcExpo8       = EEPROM.read(adr++);
    accStrength8  = EEPROM.read(adr++);
    rollPitchRate = EEPROM.read(adr++);
    yawRate       = EEPROM.read(adr++);
    dynThrPID     = EEPROM.read(adr++);

    serialWrite();
  }
  return succ;
}

void storeProfile(char* name) {
  uint16_t adr = constrain((profile+1), 1, 16) * 32;

  int i = 0;
  while (name[i] != '\0' && i < 8) {
    EEPROM.write(adr++, name[i++]);
  }
  while (i < 8) {
    name[i++] = ' ';
  }
    
  EEPROM.write(adr++, 'C');
  for (int8_t i=0; i<3; i++) {EEPROM.write(adr++, P8[i]); EEPROM.write(adr++, I8[i]); EEPROM.write(adr++, D8[i]);}
  EEPROM.write(adr++, rcRate8      );
  EEPROM.write(adr++, rcExpo8      );
  EEPROM.write(adr++, accStrength8 );
  EEPROM.write(adr++, rollPitchRate);
  EEPROM.write(adr++, yawRate      );
  EEPROM.write(adr++, dynThrPID    );
}

void deleteProfile() {
  uint16_t adr = constrain((profile+1), 1, 16) * 32;
  for (int8_t i = 0; i < 32; i++) {
    EEPROM.write(adr++, 0);
  }
}

bool getProfileName(char* name) {
  uint16_t adr = constrain((profile+1), 1, 16) * 32;
  for (int8_t i = 0; i < 8; i++) { name[i] = EEPROM.read(adr++); }
  //name[8] = '\0'; 

  char    hdr  = EEPROM.read(adr++);
  boolean succ = (hdr == 'C');  
  if (!succ) {
    for (int8_t i = 0; i < 8; i++) { name[i] = '-'; }
  }
  return succ;
} 

// ----------------

boolean restoreParams() {
  byte    magic0 = EEPROM.read(0); // check validity of data by storing some magic numbers as a prefix
  byte    magic1 = EEPROM.read(1);
  uint8_t tv     = EEPROM.read(2);
  if (tv >= 15 && tv < 16) {
    tv -= 15;
  }
  boolean succ = (magic0 == 47 && magic1 == 11 && tv >= 0 && tv < sizeof(frameLength));

  if (succ) {
    vers    = tv;
#ifdef LOGGING
    logMode   = EEPROM.read(3);
    if (logMode < 0) {
      logMode = 0;
    }
    logDelay  = EEPROM.read(4);
#endif
    backlight = 255 - EEPROM.read(5);
  }
  return succ;
}

void storeParams() {
  EEPROM.write(0, 47);  
  EEPROM.write(1, 11);  
  EEPROM.write(2, vers);
#ifdef LOGGING  
  EEPROM.write(3, logMode);
  EEPROM.write(4, logDelay);
#endif
  EEPROM.write(5, 255 - backlight);
}

// ----------------

#ifdef LOGGING

void formatLogFile(int cnt) {
  fileName[2] = '0' + cnt % 10; cnt /= 10;
  fileName[1] = '0' + cnt % 10; cnt /= 10;
  fileName[0] = '0' + cnt % 10;
}

void openLog() {
  boolean created = false; 
  int     cnt     = 0;
  while (!created && cnt < 1000) {
    formatLogFile(cnt);
    if (!SD.exists(fileName)) {
      logFile = SD.open(fileName, FILE_WRITE);
      setBacklight();
      created = true;
    }
    else {
      cnt++;
    }
  }
  if (!created) {
    logMode = 0;
    storeParams();
    setError(fileError);
  }
  state   = 0;
  logMode = 1;
  storeParams();

  lcdPrintProg(0, 0, logMsg);
  lcdPrintProg(0, 1, fileMsg);
  format(6, 1, cnt, 3, 0);
  lcd.print(" #");

  green(true);
}

int findNextLogFile(int act, boolean ascend) {
  int     cnt   = (ascend)? (act + 1) % 1000 : (act + 999) % 1000;
  boolean found = false;
  while (!found && cnt != act) {
    formatLogFile(cnt);
    found = SD.exists(fileName);
    setBacklight();
    if (!found) {
      cnt = (ascend)? (cnt + 1) % 1000 : (cnt + 999) % 1000;
    }
  }
  if (!found) {
    formatLogFile(act);          // check if exactly one logfile exists
    found = SD.exists(fileName);
    setBacklight();
  }
  return (!found)? -1 : cnt;
}

void doLog() {
  logFile.write(buffer, ptr);
  setBacklight();

  lineNo++;
  
  int8_t key = getKeyPressed();
  if (key >= 0) {
    bool flag = confirm(stopLogMsg);
    if (flag) {
      // close the logfile and stop logging
      logFile.close();
      setBacklight();
      green(false);
      lineNo  = 0;
      logMode = 0;
      storeParams();
      waitForKeyUp();

      state = 42;
      lcdPrintProg(0, 0, msgs[state]);
      lcdPrintProg(0, 1, emptyLine);
      return;
    }
    else {
      // continue logging
      lcdPrintProg(0, 1, fileMsg);
      lcd.setCursor(6, 1);
      lcd.write(fileName[0]);
      lcd.write(fileName[1]);
      lcd.write(fileName[2]);
      lcd.print(" #");
    }
  }
  int readBytes = serialRead(false);
  if (readBytes > 0) {
    format(11, 1, lineNo, 5, 0);
    blinkLed(YELLOW_LED, 10);
    delay(logDelay * 10);
  }
  else {
    logMode = 0;
    storeParams();
  }
}

boolean playbackLogFile() {
  // check if a logfile exists
  int cnt = findNextLogFile(999, true);
  boolean result = (cnt >= 0);
  if (result) {
    while (true) {
      bool playback = confirm(playMsg1);
      if (playback) {
        green(true);

        lcdPrintProg(0, 0, playMsg3);
        lcdPrintProg(0, 1, emptyLine);
        lcdPrintProg(3, 1, fileMsg);
        
        uint8_t key = -1;
        while (true) {
          format(9, 1, cnt, 3, 0);
  
          key = enterKey();
          if (key == KEY_UP || key == KEY_DOWN) {
            cnt = findNextLogFile(cnt, key == KEY_DOWN); 
          }
          else {
            break;  
          }
        }
        if (key == KEY_RIGHT || key == KEY_SELECT) {
          lcdPrintProg(0, 0, playMsg4);
          lcdPrintProg(0, 1, fileMsg);
          lcdPrintProg(6, 1, fileSuff);
          format(6, 1, cnt, 3, 0);
          
          lineNo = 0;
          formatLogFile(cnt);
          logFile = SD.open(fileName, FILE_READ);

          uint8_t len = frameLength[vers];
          while (logFile.available()) {
            format(11, 1, lineNo, 5, 0);
            lineNo++;

            ptr = 0;
            while (ptr < len && logFile.available()) {
              buffer[ptr++] = logFile.read();
            }
            setBacklight();
            buffer[ptr++] = 'A';
            if (ptr == len+1) {
              int timeout = 600;
              while (timeout > 0 && !Serial.available()) {
                if (timeout <= 590) {
                  blinkLed(RED_LED, 4);
                }
                delay(50);
                timeout--;
              }
              if (timeout <= 0) {
                lcdPrintProg(0, 1, timeoutError);
                blinkLed(RED_LED, 500, 3);
                break;
              }                
              if (Serial.read() == 'A') {
                blinkLed(YELLOW_LED, 4);
                Serial.write(buffer, ptr);
                delay(logDelay * 10); // in real-time
              }
            }
            else {
              break; // eof
            }
          }
          logFile.close();
        }
        green(false);
      }
      else {
        break; // leave
      }
    }
  }
  return result;
}

void deleteLogFile() {
  int cnt  = 999;
  int next = findNextLogFile(cnt, true);
  while (next >= 0) {
    cnt = next;
    lcdPrintProg(3, 1, fileMsg);
    format(9, 1, cnt, 3, 0);
    
    int8_t key = enterKey();
    if (key == KEY_UP || key == KEY_DOWN) {
      next = findNextLogFile(cnt, key == KEY_DOWN);
    }
    else if (key == KEY_SELECT || key == KEY_RIGHT) {
      formatLogFile(cnt);
      if (SD.remove(fileName)) {
        lcdPrintProg(0, 1, fileDeleted);
        next = findNextLogFile(cnt, true);
      }
      else {
        lcdPrintProg(0, 1, deleteFailed);
      }
      setBacklight();
      delay(1000);
      lcdPrintProg(0, 1, emptyLine);
    }
    else if (key == KEY_LEFT) {
      next = -1;
    }
  }
}
#endif

